// File: ReadersWritersC.cpp
// Readers and writers with a Mesa semantics monitor
// The Terekhov algorithm, optimized
// Stan Warford
// December 2020

#include <cstdlib>
#include <iostream>
#include <thread>
#include <mutex>
#include "Util450.cpp"

using namespace std;

class RWMonitor {
private:
    mutex rwMutex;
    condition_variable gate1;
    condition_variable gate2;
    unsigned state = 0;
    static const unsigned writerMask = 1U << (sizeof(unsigned) * CHAR_BIT - 1);
    static const unsigned readerMask = ~writerMask;

public:
    void startRead() {
        unique_lock<mutex> guard(rwMutex);
        while ((state & writerMask) || (state & readerMask) == readerMask)
            gate1.wait(guard);
        unsigned readers = (state & readerMask) + 1;
        state &= writerMask;
        state |= readers;
    }

    void endRead() {
        unique_lock<mutex> guard(rwMutex);
        unsigned readers = (state & readerMask) - 1;
        state &= writerMask;
        state |= readers;
        if (state & writerMask) {
            if (readers == 0)
                gate2.notify_one();
        } else {
            if (readers == readerMask - 1)
                gate1.notify_one();
        }
    }

    void startWrite() {
        unique_lock<mutex> guard(rwMutex);
        while (state & writerMask)
            gate1.wait(guard);
        state |= writerMask;
        while (state & readerMask)
            gate2.wait(guard);
    }

    void endWrite() {
        unique_lock<mutex> guard(rwMutex);
        state = 0;
        gate1.notify_all();
    }
};

class RWDataBase {
private:
    RWMonitor rwMonitor;
    int myData = 5;
    mutex coutMutex;

public:
    void readMyData(int readerID) {
        rwMonitor.startRead();
        coutMutex.lock();
        cout << "Reader " << readerID << " is about to read" << endl;
        coutMutex.unlock();
        randomDelay(60);
        coutMutex.lock();
        cout << "Reader " << readerID << " read " << myData << endl;
        coutMutex.unlock();
        rwMonitor.endRead();
    }

    void writeMyData(int writerID) {
        rwMonitor.startWrite();
        cout << "Writer " << writerID << " is about to write" << endl;
        randomDelay(60);
        myData += 5;
        cout << "Writer " << writerID << " wrote " << myData << endl;
        rwMonitor.endWrite();
    }
};

RWDataBase rwDataBase;

void readerRun(int readerID) {
    for (int i = 0; i < 3; i++) {
        randomDelay(60);
        rwDataBase.readMyData(readerID);
    }
}

void writerRun(int writerID) {
    for (int i = 0; i < 3; i++) {
        randomDelay(60);
        rwDataBase.writeMyData(writerID);
    }
}

int main() {
    thread reader0(readerRun, 0);
    thread reader1(readerRun, 1);
    thread reader2(readerRun, 2);
    thread writer0(writerRun, 0);
    thread writer1(writerRun, 1);
    reader0.join();
    reader1.join();
    reader2.join();
    writer0.join();
    writer1.join();
    return EXIT_SUCCESS;
}
